/*******************************************************************************\
 * Copyright (c) 1997 by Markus Koelblin
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
\*******************************************************************************/

/*******************************************************************************\
 * xcolsel.c                                                                   *
 *                                                                             *
 * Implements a rgb color selectior and rgb.txt name lookup tool for           *
 * PseudoColor (r/w colormaps) and TrueColor visuals.                          *
 *                                                                             *
 * Author:  Markus Koelblin (Markus.Koelblin@hadiko.de) 1997                   *
\*******************************************************************************/

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>

/*
 * If scrollbars don't work correctly on 32 bit machines
 * make sure -DNARROWPROTO is defined
 */
 
#include <X11/Xaw/Scrollbar.h>

#include <stdio.h>
#include "colors.h"
#include "rgbtxt.h"
#include "globals.h"

#include "bitmaps/xcolsel.xbm"
#include "bitmaps/check.xbm"

#ifdef BIG_BITMAPS
#include "bitmaps/clear.xbm"
#include "bitmaps/look.xbm"
#else
#include "bitmaps/clear16.xbm"
#include "bitmaps/look16.xbm"
#endif

/*******************************************************************************\
 * Macros                                                                      *
\*******************************************************************************/

#define APP_CLASS_NAME   "Xcolsel"

#define RGBFORM_WIDTH    192
#define RGBFORM_HEIGHT   100
#define COLOR_WIDTH      128
#define COLOR_HEIGHT     64
#define TEXT_WIDTH       112

#define SCROLL_MAX       256
#define SCROLL_LIMIT(pos) ((pos < 0) ? 0 : ((pos < SCROLL_MAX) ? pos : SCROLL_MAX-1))


/*******************************************************************************\
 * Typedefs                                                                    *
\*******************************************************************************/

typedef struct RgbFormStruct
{
    Widget    showcolor;                     /* Accessable RgbForm Widgets */
    Widget    red_scroll,  red_label;
    Widget    green_scroll, green_label;
    Widget    blue_scroll, blue_label;
    short     red, green, blue;              /* values changed by scrollbars */
    Pixel     pixel;                         /* pixel index into colormap */
}
RgbFormData;

typedef struct MenuFormStruct
{
    Widget    color;
    Widget    menu;
    Widget    name;
    Widget    clear;
    Widget    text;
}
MenuFormData;

typedef struct XcselStruct
{
    MenuFormData  menuform;
    RgbFormData   rgbform;
    RgbQuery      query_idx;
    RgbBase       color_db;
}
XcselData;


/*******************************************************************************\
 * Global data                                                                 *
\*******************************************************************************/

XtAppContext  app_context;
XVisualInfo   visual_info;
Widget        toplevel;
Atom          wm_delete_window;
Atom          wm_protocols;
Pixmap        xcsel_icons[ICON_COUNT] = { 0, 0, 0, 0, };
Colormap      colormap = 0;

/* Application specific data */
XcselData     appdata;
Boolean       rw_colormap;
Boolean       exact_lookup = False;

/*******************************************************************************\
 * Global X-Toolkit specific resources                                         *
\*******************************************************************************/

struct OptionVals
{
    char  *rgb_file;
    short tolerance;
}
options;

XrmOptionDescRec cmd_options[] =
{
    {"-near",    ".Near",    XrmoptionSepArg, NULL},
    {"-rgbfile", ".rgbFile", XrmoptionSepArg, NULL},
};

XtResource resources[] =
{
    { "rgbFile", "RgbFile", XtRString, sizeof (char*),
      XtOffsetOf(struct OptionVals, rgb_file), XtRString, RGBTXT_DB_NAME },
    { "near", "Near", XtRShort, sizeof(int),
      XtOffsetOf(struct OptionVals, tolerance), XtRImmediate, (XtPointer) 20 },
};

String fallback_resources[] =
{
    "*menupane.preferredPaneSize: 56",
    "*color.label:           Color",
    "*color.leftBitmap:      menu12",
    "*menu*leftMargin:       16",
    "*menu*rightMargin:      16",
    "*menu.list.label:       List colors",
    "*menu.query.label:      Query colors",
    "*menu.lookup.label:     Match name",
    "*menu.exact.label:      Exact mode",
    "*menu.quit.label:       Exit",
    "*sb_form.red.label:     r:",
    "*sb_form.green.label:   g:",
    "*sb_form.blue.label:    b:",
    "*sb_form.borderWidth:   0",
    "*colorform.borderWidth: 0",
    "*showcolor.label:        ",
    NULL,
};


/*******************************************************************************\
 * set_rgb_colorbutton()                                                       *
 *                                                                             *
 * Update rgbform's color button - used by the callbacks                       *
\*******************************************************************************/

static void set_rgb_buttoncolor(RgbFormData* data)
{
    Display *display = XtDisplay(toplevel);
    XColor  color;

    color.red   = SCROLL_MAX * data->red;
    color.green = SCROLL_MAX * data->green;
    color.blue  = SCROLL_MAX * data->blue;
    color.flags = DoRed | DoGreen | DoBlue;

    if (rw_colormap)
    {
	color.pixel = data->pixel;
	XStoreColor(display, colormap, &color);
    }
    else
    {
	XAllocColor(display, colormap, &color);
	data->pixel = color.pixel;
    }
    XtVaSetValues(data->showcolor, XtNbackground, data->pixel, NULL);
}


/*******************************************************************************\
 * set_rgb_scrollbars()                                                        *
 *                                                                             *
 * Set initial scrollbar positions and labels for the RgbForm                  *
\*******************************************************************************/

static void set_rgb_scrollbars(RgbFormData* data)
{
    Dimension length;
    char      buffer[8];

    /* Initialize red scrollbar */
    XtVaGetValues(data->red_scroll, XtNlength, &length, NULL);
    XawScrollbarSetThumb(data->red_scroll, (float)data->red/(float)SCROLL_MAX, -1.0F);
    sprintf(buffer, "%3d", data->red);
    XtVaSetValues(data->red_label, XtNlabel, buffer, NULL);

    /* Initialize green scrollbar */
    XtVaGetValues(data->green_scroll, XtNlength, &length, NULL);
    XawScrollbarSetThumb(data->green_scroll, (float)data->green/(float)SCROLL_MAX, -1.0F);
    sprintf(buffer, "%3d", data->green);
    XtVaSetValues(data->green_label, XtNlabel, buffer, NULL);

    /* Initialize blue scrollbar */
    XtVaGetValues(data->blue_scroll, XtNlength, &length, NULL);
    XawScrollbarSetThumb(data->blue_scroll, (float)data->blue/(float)SCROLL_MAX, -1.0F);
    sprintf(buffer, "%3d", data->blue);
    XtVaSetValues(data->blue_label, XtNlabel, buffer, NULL);
}


/*******************************************************************************\
 * show_listbox_selection()                                                    *
 *                                                                             *
 * Called from listdlg.c to show selected color in the xcsel window.           *
\*******************************************************************************/

void show_listbox_selection(RgbBase* base, int index)
{
    RgbTxtEntry *entries = base->entries;

    if (index >= 0 && index < base->count)
    {
	appdata.rgbform.red   = entries[index].red;
	appdata.rgbform.green = entries[index].green;
	appdata.rgbform.blue  = entries[index].blue;
	XtVaSetValues(appdata.menuform.text, XtNstring, entries[index].name, NULL);

	set_rgb_buttoncolor(&appdata.rgbform);
	set_rgb_scrollbars(&appdata.rgbform);
    }
}


/*******************************************************************************\
 *                                                                             * 
 * Callbacks                                                                   *
 *                                                                             * 
\*******************************************************************************/

void quit_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData *data    = (XcselData*) client_data;
    Display   *display = XtDisplay(toplevel);

    /* Free X resources */
    if (colormap)   XFreeColormap(display, colormap);
    if (ICON_MAIN)  XFreePixmap(display, ICON_MAIN);
    if (ICON_CHECK) XFreePixmap(display, ICON_CHECK);
    if (ICON_LOOK)  XFreePixmap(display, ICON_LOOK);
    if (ICON_CLEAR) XFreePixmap(display, ICON_CLEAR);

    /* Free application resources */
    rgbtxt_free_db(&data->color_db);
    exit(0);
}

void match_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData*  data  = (XcselData*) client_data;
    Boolean     match = False;
    String      result;
    int         length;

    XtVaGetValues(data->menuform.text, XtNstring, &result, NULL);

    if (*result == '#' && strlen(result) == 7)
    {
	char  dummy;
	int   nread;

	nread = sscanf(result, "%c%02x%02x%02x", &dummy,
		       &data->rgbform.red,
		       &data->rgbform.green,
		       &data->rgbform.blue);
	match = (nread == 4);
    }
    else if (*result != '\0')
    {
	XColor color;

	if (XParseColor(XtDisplay(toplevel), colormap, result, &color))
	{
	    data->rgbform.red   = color.red   / SCROLL_MAX;
	    data->rgbform.green = color.green / SCROLL_MAX;
	    data->rgbform.blue  = color.blue  / SCROLL_MAX;
	    match = True;
	}
    }

    if (match)
    {
	set_rgb_buttoncolor(&data->rgbform);
	set_rgb_scrollbars(&data->rgbform);
    }
    else
	XBell(XtDisplay(toplevel), 100);
}

void color_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData*  data= (XcselData*) client_data;
    static char strbuf[32];

    sprintf(strbuf, "#%02x%02x%02x", 
	    (unsigned char) data->rgbform.red,
	    (unsigned char) data->rgbform.green,
	    (unsigned char) data->rgbform.blue);

    XtVaSetValues(data->menuform.text, XtNstring, strbuf, NULL);
}

void name_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData*  data  = (XcselData*) client_data;
    Boolean     match = False;
    String      result;

    XtVaGetValues(data->menuform.text, XtNstring, &result, NULL);

    if (*result == '#' && strlen(result) == 7)
    {
	int red, green, blue;
	int nread;

	nread = sscanf(result+1, "%02x%02x%02x", &red, &green, &blue);

	if (nread == 3)
	{
	    if (exact_lookup)
		match = rgbtxt_match_name(&data->color_db, &result, red, green, blue);
	    else
		match = rgbtxt_guess_name(&data->color_db, &result, red, green, blue,
					  options.tolerance);
	}
    }

    if (match)
	XtVaSetValues(data->menuform.text, XtNstring, result, NULL);
    else
	XBell(XtDisplay(toplevel), 100);
}

void clear_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData*  data = (XcselData*) client_data;
    static char strbuf[1] = "";

    XtVaSetValues(data->menuform.text, XtNstring, strbuf, NULL);
}


void exact_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    static Pixmap old_bm = 0;

    if (exact_lookup)
    {
	XtVaSetValues(w, XtNleftBitmap, old_bm, NULL);
	exact_lookup = False;
    }
    else
    {
	if (!old_bm) XtVaGetValues(w, XtNleftBitmap, &old_bm, NULL);
	XtVaSetValues(w, XtNleftBitmap, ICON_CHECK, NULL);
	exact_lookup = True;
    }
}

void query_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData   *data = (XcselData*) client_data;
    RgbQuery    *query;

    /* Create new query structure */
    if ((query  = (RgbQuery*) malloc(sizeof(RgbQuery))) == NULL)
	return;

    /* Setup pattern and tolerance for query */
    query->pattern.red   = data->rgbform.red;
    query->pattern.green = data->rgbform.green;
    query->pattern.blue  = data->rgbform.blue;
    sprintf(query->pattern.name, "#%02x%02x%02x",
	    (unsigned char) data->rgbform.red,
	    (unsigned char) data->rgbform.green,
	    (unsigned char) data->rgbform.blue);
    query->tolerance = options.tolerance;

    /*
     * If there are matches create index list and show the result.
     * Note: The memory for query structure and index list is
     * disposed by colorlist_dialog's close callback.
     */
    if (rgbtxt_create_query(&data->color_db, query))
	colorlist_dialog(w, colormap, &data->color_db, query);
}

void list_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XcselData*  data = (XcselData*) client_data;

    colorlist_dialog(w, colormap, &data->color_db, NULL);
}


/*******************************************************************************\
 *                                                                             *
 * Callbacks for all the RGB Scrollbars                                        *
 *                                                                             *
\*******************************************************************************/

static void rgb_scroll_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    RgbFormData* data   = (RgbFormData*) client_data;
    const long   pixels = (long) call_data;
    Widget       label;
    short        *value;
    int          pos;

    /* Determine the calling scrollbar */
    if (w == data->red_scroll)
    {
	value = &data->red;
	label = data->red_label;
    }
    else if (w == data->green_scroll)
    {
	value = &data->green;
	label = data->green_label;
    }
    else
    {
	value = &data->blue;
	label = data->blue_label;
    }

    /* If pixels is negative, left button, else right button pressed. */
    if (pixels < 0)
	pos = SCROLL_LIMIT(*value-1);
    else
	pos = SCROLL_LIMIT(*value+1);

    if (pos != *value)
    {
	Dimension length;
	char      buffer[8];

	XtVaGetValues(w, XtNlength, &length, NULL);
	XawScrollbarSetThumb(w, (float)pos/(float)SCROLL_MAX, (float)length/(100.F*SCROLL_MAX));
	sprintf(buffer, "%3d", pos);
	XtVaSetValues(label, XtNlabel, buffer, NULL);
	*value = pos;
	set_rgb_buttoncolor(data);
	color_CB(w, (XtPointer) &appdata , NULL);
    }
}

static void rgb_thumb_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
    RgbFormData* data    = (RgbFormData*) client_data;
    const float  percent = *((float*) call_data);
    const int    pos     = SCROLL_LIMIT( (int)(SCROLL_MAX*percent) );
    Widget       label;
    short        *value;

    /* Determine the calling scrollbar (duplicated code...) */
    if (w == data->red_scroll)
    {
	value = &data->red;
	label = data->red_label;
    }
    else if (w == data->green_scroll)
    {
	value = &data->green;
	label = data->green_label;
    }
    else
    {
	value = &data->blue;
	label = data->blue_label;
    }

    if (pos != *value)
    {
	char buffer[8];

	XawScrollbarSetThumb(w, (float)pos/(float)SCROLL_MAX, -1.0F);
	sprintf(buffer, "%3d", pos);
	XtVaSetValues(label, XtNlabel, buffer, NULL);
	*value = pos;
	set_rgb_buttoncolor(data);
	color_CB(w, (XtPointer) &appdata , NULL);
    }
}


/*******************************************************************************\
 * create_icons()                                                              *
 *                                                                             *
 * Create all application icons and store them in the global xcsel_icons array *
 *******************************************************************************/

static void create_icons(Display *display, Window window)
{
    ICON_MAIN  = XCreateBitmapFromData(display, window, (char*) xcolsel_bits,
				       xcolsel_width, xcolsel_height);
    ICON_CHECK = XCreateBitmapFromData(display, window, (char*) check_bits,
				       check_width, check_height);
    ICON_LOOK  = XCreateBitmapFromData(display, window, (char*) look_bits,
				       look_width, look_height);
    ICON_CLEAR = XCreateBitmapFromData(display, window, (char*) clear_bits,
				       clear_width, clear_height);
}


/*******************************************************************************\
 * create_menu_entries()                                                       *
 *                                                                             *
 * Create menu entries and add callbacks for the color menu.                   *
 *******************************************************************************/

static void create_menu_entries(Widget menu)
{
    Widget entry;

    entry = XtCreateManagedWidget("list", smeBSBObjectClass, menu, NULL, 0);
    XtAddCallback(entry, XtNcallback, list_CB, (XtPointer) &appdata);
    entry = XtCreateManagedWidget("query", smeBSBObjectClass, menu, NULL, 0);
    XtAddCallback(entry, XtNcallback, query_CB, (XtPointer) &appdata);
    entry = XtCreateManagedWidget("lookup", smeBSBObjectClass, menu, NULL, 0);
    XtAddCallback(entry, XtNcallback, name_CB, (XtPointer) &appdata);
    entry = XtCreateManagedWidget("exact", smeBSBObjectClass, menu, NULL, 0);
    XtAddCallback(entry, XtNcallback, exact_CB, (XtPointer) &appdata);
    entry = XtCreateManagedWidget("line", smeLineObjectClass, menu, NULL, 0);
    entry = XtCreateManagedWidget("quit", smeBSBObjectClass, menu, NULL, 0);
    XtAddCallback(entry, XtNcallback, quit_CB, (XtPointer) &appdata);
}


/*******************************************************************************\
 * create_menuform()                                                           *
 *                                                                             *
 * Create the Widget tree with menubuttons                                     *
\*******************************************************************************/

static void create_menuform(Widget parent, MenuFormData* data)
{
    Widget menupane, menubox, toolbox;
    String text_trans = "#override\n\
                         <Key>Return: match-action()\n\
                         <Key>Linefeed: no-op()\n\
                         <Key>Delete: delete-next-character()\n\
                         Ctrl<Key>s: no-op()\n\
                         Ctrl<Key>r: no-op()";

    /* Create pane and insert menubox and toolbox */
    menupane    = XtCreateManagedWidget("menupane", panedWidgetClass, parent, NULL, 0);
    menubox     = XtCreateManagedWidget("menubox",  boxWidgetClass, menupane, NULL, 0);
    toolbox     = XtCreateManagedWidget("toolbox",  boxWidgetClass, menupane, NULL, 0);

    /* Create menu button and popup, fill popup with entries  */
    data->color = XtCreateManagedWidget("color", menuButtonWidgetClass, menubox, NULL, 0);
    data->menu  = XtCreatePopupShell("menu", simpleMenuWidgetClass, data->color, NULL, 0);
    create_menu_entries(data->menu);
    XtVaSetValues(data->color, XtNmenuName, "menu",  NULL);

    /* Textinput and speedbuttons for the toolbox */ 
    data->text  = XtVaCreateManagedWidget("text", asciiTextWidgetClass, toolbox,
					  XtNtranslations, XtParseTranslationTable(text_trans),
					  XtNeditType, XawtextEdit,
					  XtNresize, XawtextResizeWidth,
					  XtNwidth, TEXT_WIDTH,
					  XtNwrap, XawtextWrapNever,
					  NULL);
    data->name  = XtVaCreateManagedWidget("name", commandWidgetClass, toolbox,
					  XtNlabel, "  ",
					  NULL);
    data->clear = XtVaCreateManagedWidget("clear", commandWidgetClass, toolbox,
					  XtNlabel, "  ",
					  NULL);

    /* Add callbacks for the menu widgets */
    XtAddCallback(data->name,  XtNcallback, name_CB,  (XtPointer) &appdata);
    XtAddCallback(data->clear, XtNcallback, clear_CB, (XtPointer) &appdata);

    /* All keyboard inputs goes to menuform's text widget */
    XtSetKeyboardFocus(parent, data->text);
}

/*******************************************************************************\
 * create_rgbform()                                                             *
 *                                                                             *
 * Create the Widget tree for color selector                                   *
\*******************************************************************************/

static void create_rgbform(Widget parent, RgbFormData* data)
{
    Widget rgbform, sb_form, colorform, tmplabel;

    rgbform = XtVaCreateManagedWidget("rgbform", formWidgetClass, parent, NULL);

    /*
     *  Put the color button in a seperate subform
     */
    colorform = XtVaCreateManagedWidget(
	"colorform", formWidgetClass, rgbform,
	XtNright,    XtChainLeft,
	XtNleft,     XtChainLeft,
	XtNtop,      XtChainTop,
	XtNbottom,   XtChainTop,
	XtNwidth,    RGBFORM_WIDTH,
	XtNheight,   RGBFORM_HEIGHT,
	NULL);

    data->showcolor = XtVaCreateManagedWidget(
	"showcolor", commandWidgetClass, colorform,
	XtNwidth,    COLOR_WIDTH,
	XtNheight,   COLOR_HEIGHT,
	NULL);
    /*
     *  Put scrollbars in a seperate subform
     */
    sb_form  = XtVaCreateManagedWidget(
	"sb_form",  formWidgetClass, rgbform,
	XtNright,   XtChainLeft,
	XtNleft,    XtChainLeft,
	XtNtop,     XtChainTop,
	XtNbottom,  XtChainTop,
	NULL);

    /*
     *  Red scrollbar stuff
     */
    tmplabel = XtVaCreateManagedWidget(
	"red",          labelWidgetClass, sb_form,
	XtNfromVert,    colorform,
	XtNborderWidth, 0,
	NULL);

    data->red_label = XtVaCreateManagedWidget(
	"red_label",     labelWidgetClass, sb_form,
	XtNfromVert,    colorform,
	XtNfromHoriz,   tmplabel,
	XtNresizable,   False,
	XtNborderWidth, 0,
	NULL);

    data->red_scroll = XtVaCreateManagedWidget(
	"red_scroll", scrollbarWidgetClass, sb_form,
	XtNfromVert,    data->red_label,
	XtNorientation, XtorientHorizontal,
	XtNright,       XtChainRight,
	XtNleft,        XtChainLeft,
	XtNtop,         XtChainTop,
	XtNbottom,      XtChainTop,
	XtNwidth,       RGBFORM_WIDTH,
	NULL);
    /*
     *  Green scrollbar stuff
     */
    tmplabel = XtVaCreateManagedWidget(
	"green",        labelWidgetClass, sb_form,
	XtNfromVert,    data->red_scroll,
	XtNborderWidth, 0,
	NULL);

    data->green_label = XtVaCreateManagedWidget(
	"green_label",   labelWidgetClass, sb_form,
	XtNfromVert,    data->red_scroll,
	XtNfromHoriz,   tmplabel,
	XtNresizable,   True,
	XtNborderWidth, 0,
	NULL);

    data->green_scroll= XtVaCreateManagedWidget(
	"green_scroll",  scrollbarWidgetClass, sb_form,
	XtNfromVert,    data->green_label,
	XtNorientation, XtorientHorizontal,
	XtNright,       XtChainRight,
	XtNleft,        XtChainLeft,
	XtNtop,         XtChainTop,
	XtNbottom,      XtChainTop,
	XtNwidth,       RGBFORM_WIDTH,
	NULL);
    /*
     *  Blue scrollbar stuff
     */
    tmplabel = XtVaCreateManagedWidget(
	"blue",         labelWidgetClass, sb_form,
	XtNfromVert,    data->green_scroll,
	XtNborderWidth, 0,
	NULL);

    data->blue_label = XtVaCreateManagedWidget(
	"blue_label",    labelWidgetClass, sb_form,
	XtNfromVert,    data->green_scroll,
	XtNfromHoriz,   tmplabel,
	XtNresizable,   True,
	XtNborderWidth, 0,
	NULL);

    data->blue_scroll= XtVaCreateManagedWidget(
	"blue_scroll", scrollbarWidgetClass, sb_form,
	XtNfromVert,     data->blue_label,
	XtNorientation,  XtorientHorizontal,
	XtNright,        XtChainRight,
	XtNleft,         XtChainLeft,
	XtNtop,          XtChainTop,
	XtNbottom,       XtChainTop,
	XtNwidth,        RGBFORM_WIDTH,
	NULL);

    /* Add callbacks for RGB scrollbars */
    XtAddCallback(data->red_scroll,   XtNscrollProc, rgb_scroll_CB, (XtPointer) data);
    XtAddCallback(data->red_scroll,   XtNjumpProc,   rgb_thumb_CB,  (XtPointer) data);
    XtAddCallback(data->green_scroll, XtNscrollProc, rgb_scroll_CB, (XtPointer) data);
    XtAddCallback(data->green_scroll, XtNjumpProc,   rgb_thumb_CB,  (XtPointer) data);
    XtAddCallback(data->blue_scroll,  XtNscrollProc, rgb_scroll_CB, (XtPointer) data);
    XtAddCallback(data->blue_scroll,  XtNjumpProc,   rgb_thumb_CB,  (XtPointer) data);
    /* Add callback for the color button */
    XtAddCallback(data->showcolor, XtNcallback, color_CB, (XtPointer) &appdata);
}


/*************************************************************************\
 *                                                                       *
 *  Window actions                                                       *
 *                                                                       *
\*************************************************************************/

static void match_AC(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* Match color name on menuform.text <Key>: Return */ 
    match_CB(w, (XtPointer) &appdata, NULL);
}

static void quit_AC(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    if(event->type == ClientMessage && event->xclient.type == wm_protocols)
	if (event->xclient.data.l[0] != wm_delete_window)
	    return;
    /* Free resources if applicatiobs is terminated by the WindowManager */
    quit_CB(w, (XtPointer) &appdata, NULL);
}

/* Define actions list for XtAppAddActions() */

static XtActionsRec window_actions[] =
{
    { "match-action", match_AC },
    { "quit-action",  quit_AC  },
};


/*******************************************************************************\
 *  main()                                                                     *
 *                                                                             *
 *  xcolsel RGB color selector and lookup program                              *
 *******************************************************************************/

void main(int argc, char **argv)
{
    String  wm_trans = "<ClientMessage>WM_PROTOCOLS: quit-action()\n";
    Widget  mainpane;
    XColor  color;
    Pixel   pixels[1];
    int     screen_nr;

    /* Initialize the application */
    toplevel = XtAppInitialize(&app_context, APP_CLASS_NAME,
			       cmd_options, XtNumber(cmd_options),
			       &argc, argv,
			       fallback_resources,
			       NULL, 0);

    XtGetApplicationResources(toplevel, (XtPointer) &options, resources,
			      XtNumber(resources), NULL, 0);

    XtAppAddActions(app_context, window_actions, XtNumber(window_actions));

    /* Create interface */
    mainpane = XtCreateManagedWidget("mainpane", panedWidgetClass, toplevel, NULL, 0);
    create_menuform(mainpane, &appdata.menuform);
    create_rgbform(mainpane,  &appdata.rgbform);

    /* Create color database for rgb to name conversion */
    if (!rgbtxt_create_db(&appdata.color_db, options.rgb_file))
	fputs("Warning: Not all color entries could be read.\n", stderr);

    screen_nr = DefaultScreen(XtDisplay(toplevel));

    if (XMatchVisualInfo(XtDisplay(toplevel), screen_nr, 16, TrueColor, &visual_info) ||
	XMatchVisualInfo(XtDisplay(toplevel), screen_nr, 24, TrueColor, &visual_info))
    {
	colormap = DefaultColormap(XtDisplay(toplevel), screen_nr);
	rw_colormap = False;
	set_xcolor_rgb(&color, RGBVAL_MAX/2, RGBVAL_MAX/2, RGBVAL_MAX/2);
	XAllocColor(XtDisplay(toplevel), colormap, &color);
    }
    else
    {
	/* Allocate a read/write colormap with existing entries copied */
	create_colormap(XtDisplay(toplevel), &colormap, pixels, 1);
	rw_colormap = True;

	/* Setup initial color using an r/w pixel */
	set_xcolor_rgb(&color, RGBVAL_MAX/2, RGBVAL_MAX/2, RGBVAL_MAX/2);
	color.pixel = pixels[0];
	XStoreColor(XtDisplay(toplevel), colormap, &color);
    }

    /* Initialize rgbform data, scrollbar positions and label color */
    appdata.rgbform.red   = color.red   / SCROLL_MAX;
    appdata.rgbform.green = color.green / SCROLL_MAX;
    appdata.rgbform.blue  = color.blue  / SCROLL_MAX;
    appdata.rgbform.pixel = color.pixel;
    XtVaSetValues(appdata.rgbform.showcolor, XtNbackground, color.pixel, NULL);
    set_rgb_scrollbars(&appdata.rgbform);
    color_CB(toplevel, (XtPointer) &appdata , NULL);

    /* Show window, install colormap and icons */
    XtSetMappedWhenManaged(toplevel, FALSE);
    XtRealizeWidget(toplevel);
    XSetWindowColormap(XtDisplay(toplevel), XtWindow(toplevel), colormap);

    /* Create application icons xcsel_icons */
    create_icons(XtDisplay(toplevel), XtWindow(toplevel));
    XtVaSetValues(appdata.menuform.name, XtNbitmap, ICON_LOOK,  NULL);
    XtVaSetValues(appdata.menuform.clear, XtNbitmap, ICON_CLEAR, NULL);
    XtVaSetValues(toplevel, XtNiconPixmap, ICON_MAIN, NULL);

    /* Setup WM_DELETE_WINDOW action handler */
    wm_protocols     = XInternAtom(XtDisplay(toplevel), "WM_PROTOCOLS", False);
    wm_delete_window = XInternAtom(XtDisplay(toplevel), "WM_DELETE_WINDOW", False);
    XtOverrideTranslations(toplevel, XtParseTranslationTable(wm_trans));
    XSetWMProtocols(XtDisplay(toplevel), XtWindow(toplevel), &wm_delete_window, 1);

    /* Map Window and enter eventloop */
    XtMapWidget(toplevel);
    XtAppMainLoop(app_context);
}

/*************************************************************************/

